// vtimer.h

#ifndef _VTIMER_H_
#define _VTIMER_H_

#include "stacklist.h"

struct vtimer_start {
   unsigned long long start : 48;
   unsigned rcount : 16; // recursion count.  if 0, timer is off.
};

struct vtimer {
   uint64_t total_field;
   struct vtimer_start start_field; // 64 bits

   // Stop Routine: update saved_timer.auxdata, vtimer.total_field using
   //               read of hardware counter & set vtimer.start_field.start = 0
   void (*stop_routine)(struct vtimer *, struct saved_timer *,
                        unsigned type);

   // Restart Routine: use aux_data.rcount and read of hardware counter to
   //                  set vtimer.start_field, and update vtimer.total_field
   //                  if counting events switched out
   void (*restart_routine)(struct vtimer *, uint64_t /*auxdata*/,
                           unsigned type, unsigned wtime);

   // Type: ((PERFCTR_NUM << 16) | [WALL_TIMER] | kapi_hwcounter_kind)
   unsigned counter_type;
};

// offsets from start of vtimer
#define total_field_offset (0)
#define start_field_offset (total_field_offset + sizeof(uint64_t))
#define stop_routine_offset (start_field_offset + sizeof(struct vtimer_start))
#define restart_routine_offset (stop_routine_offset + sizeof(void*))
#define counter_type_offset (restart_routine_offset + sizeof(void*))
#define sizeof_vtimer (sizeof(struct vtimer))

// macros for extracting various parts of counter_type
#define WALL_TIMER (1 << 15)
#define IS_WALL_TIMER(t) (t & WALL_TIMER)
#define CTR_OF(t) (t >> 16)
#define KIND_OF(t) (t & 0x00007fff)

#endif
